%{
(*
 * File pl0.y
 *
 * PL0 grammar in Yacc format by H. Weber
 *
 * Extensions to PL/0:   else-part of if-statement
 *                       repeat-statement
 *                       read-statement
 *                       write-statement
 *)

USES
   PL0INIT, PL0CODE, PL0ERR, PL0SYMT,
   PL0LEX,  PL0STOP, LexLib, YaccLib;

VAR
   arrid : alfa;
   n, i, k: INTEGER;

 PROCEDURE yymsg(msg: string);
 BEGIN
   writeln(sourcefile, ':', line_no, ': ', msg, ' at or before `', act_token, '''.')
 END (*yymsg*);

%}

/* note the Pascal keyword tokens are stropped with leading underscore
   (e.g. _AND) in the Turbo Pascal Yacc version, because these identifiers
   will be declared as token numbers in the header file PL0.h generated
   by Yacc, and hence must not collide with Turbo Pascal keywords.
   HW/AG */


%union {
   number  : INTEGER;
   ident   : alfa;
}

%left  PLUS MINUS
%left  STAR SLASH
%left  LPAREN RPAREN
%right UMINUS

%token <number> NUMBER
%token <ident> IDENTIFIER
%token <ident> _STRING
%token ASSIGNMENT
%token _BEGIN
%token CALL
%token _CASE
%token COLON
%token COMMA
%token _CONST
%token _DO
%token DOT
%token _DOWNTO
%token _ELSE
%token _END
%token EQUAL
%token _FOR
%token GREATERE
%token GREATER
%token _IF
%token LBRACKET
%token LESSE
%token LESS
%token NOTEQUAL
%token ODD
%token _OF
%token _PROCEDURE
%token RBRACKET
%token _READ
%token _REPEAT
%token SEMICOLON
%token _THEN
%token _TO
%token _UNTIL
%token _VAR
%token _WHILE
%token _WRITE
%token _WRITELN
%token UMINUS

%%
file
    : program
    | program error
        { writeln(sourcefile, ':', line_no,
            ':Text follows logical end of program.'); }
    ;

program
    : block DOT
    | block error
                      { Error(9); }
    ;

identifier_list
    : identifier_list COMMA identifier
    | identifier
    ;

identifier
    : IDENTIFIER
                      {
                        id:= $1;
                        Enter(varobj, dindex[dindexptr], tx[txptr]);
                      }
    | IDENTIFIER LBRACKET
                      {
                        id:= $1;
                        arrid:= id;
                      }
      array_size RBRACKET
                      {
                        id:= arrid;
                        Enter(arrobj, dindex[dindexptr], tx[txptr]);
                      }
    ;

array_size
    : NUMBER
                      { arrsize:= $1 }
    | IDENTIFIER
                      {
                        id:= $1;
                        i:= position(id, tx[txptr]);
                        IF i = 0 THEN
                           error(11)
                        ELSE WITH symtable[i] DO
                        CASE kind OF
                          constobj : arrsize:= Val;
                          varobj, procobj : error(46);
                        END;
                      }
    ;

block
    : empty
                      {
                        txfirst:= tx[txptr]+1;
                        INC(pposptr);
                        IF pposptr>maxstack THEN Error(32);
                        ppos[pposptr]:= tx[txptr];
                        INC(dindexptr);
                        IF dindexptr>maxstack THEN Error(32);
                        dindex[dindexptr]:= 3;
                        symtable[tx[txptr]].addr:= codeindex;
                        Generate(jpu,0,0);
                        IF lev>nestmax THEN Error(32);
                      }
      constant_definition_part
      variable_declaration_part
      procedure_declaration_part
                      {
                        code[symtable[ppos[pposptr]].addr].a:= codeindex;
                        symtable[ppos[pposptr]].addr:= codeindex;
                        symtable[ppos[pposptr]].size:= dindex[dindexptr];
                        Generate(int,0,dindex[dindexptr]);
                        DEC(pposptr);
                        DEC(dindexptr);
                      }
      statement
                      {
                        Generate(ret,0,0);
                      }
    ;

empty
    :
    ;

constant_definition_part
    : _CONST constant_list SEMICOLON
    | _CONST constant_list error
                      { Error(5); }
    |
    ;

constant_list
    : constant_list COMMA constant_definition
                      { yyerrok; }
    | constant_definition
    ;

constant_definition
    : IDENTIFIER EQUAL NUMBER
                      {
                        id:= $1;
                        num:= $3;
                        Enter(constobj, dindex[dindexptr], tx[txptr]);
                      }
    ;

variable_declaration_part
    : _VAR identifier_list SEMICOLON
    |
    ;

procedure_declaration_part
    : proc_declaration_list SEMICOLON
    |
    ;

proc_declaration_list
    : proc_declaration_list SEMICOLON proc_declaration
                      { yyerrok; }
    | proc_declaration
    ;

proc_declaration
    : procedure_head block
                      {
                        DEC(txptr);
                        DEC(lev);
                      }
    ;

procedure_head
    : procedure_ident SEMICOLON
                      {
                        INC(txptr);
                        tx[txptr]:= tx[txptr-1];
                        INC(lev);
                      }


procedure_ident
    : _PROCEDURE IDENTIFIER
                      {
                        id:= $2;
                        Enter(procobj, dindex[dindexptr], tx[txptr]);
                      }
    ;

statement
    : assignment_statement
    | case_statement
    | compound_statement
    | if_statement
    | for_statement
    | procedure_statement
    | read_statement
    | repeat_statement
    | while_statement
    | write_statement
    | writeln_statement
    |
    ;

assignment_statement
    : IDENTIFIER ASSIGNMENT expression
                      {
                        id:= $1;
                        i:= Position(id,tx[txptr]);
                        IF i=0 THEN
                          Error(11)
                        ELSE IF symtable[i].kind<>varobj THEN
                          Error(12);
                        WITH symtable[i]
                          DO Generate(sto,lev-level,addr);
                      }
    | IDENTIFIER LBRACKET
                      {
                        id:= $1;
                        i:= Position(id,tx[txptr]);
                        IF i=0 THEN
                          Error(11)
                        ELSE IF symtable[i].kind<>arrobj THEN
                          Error(49);
                        INC(arrptr);
                        IF arrptr>maxstack THEN Error(32);
                        arr[arrptr]:= i;
                      }
      expression RBRACKET
                      {
                        i:= arr[arrptr];
                        generate(lit,0,symtable[i].addr);
                        generate(add,0,0);
                      }
      ASSIGNMENT expression
                      {
                        i:= arr[arrptr];
                        generate(sta,lev-symtable[i].level,0);
                        DEC(arrptr);
                      }
    ;

case_statement
    : _CASE
                      {
                        INC(caseptr);
                        caseinfo[caseptr].i:= 0;
                        caseinfo[caseptr].j:= 0;
                      }
       expression _OF
                      {
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                        Generate(jpu,0,0);
                      }
      case_statement_list
      _END
                      {
                        code[fixup[fixupptr]].a:= codeindex;
                        FOR k:= 1 TO caseinfo[caseptr].i DO
                        BEGIN
                          Generate(lit,0,
                                   caseinfo[caseptr].casetab[k].value);
                          Generate(ldl,0,
                                   caseinfo[caseptr].casetab[k].cdindex);
                        END;
                        Generate(jpx,0,caseinfo[caseptr].i);
                        FOR k:= 1 TO caseinfo[caseptr].j DO
                          code[caseinfo[caseptr].exittab[k]].a:= codeindex;
                        DEC(fixupptr);
                        DEC(caseptr);
                      }
    ;

case_statement_list
    : case_list
    | case_list SEMICOLON
    ;

case_list
    : case_list SEMICOLON onecase
    | onecase
    ;

onecase
    : NUMBER
                      {
                        WITH caseinfo[caseptr] DO
                        BEGIN
                          IF i = casemax THEN
                            Error(37)
                          ELSE
                          BEGIN
                            INC(i);
                            num:= $1;
                            casetab[i].value:= num;
                            casetab[i].cdindex:= codeindex;
                            k:= 0;
                            REPEAT
                              INC(k)
                            UNTIL casetab[k].value = num;
                            IF k < i THEN
                              Error(40);
                          END;
                        END;
                      }
      COLON
      statement
                      {
                        INC(caseinfo[caseptr].j);
                        caseinfo[caseptr].exittab[caseinfo[caseptr].j]
                          := codeindex;
                        Generate(jpu,0,0);
                      }
    ;

compound_statement
    : _BEGIN statement_sequence _END
    ;

statement_sequence
    : statement_sequence SEMICOLON statement
    | statement
    | error
                      {
                        writeln('statement ignored');
                        yyclearin
                      }
    ;

if_statement
    : if_prefix statement
                      {
                        code[fixup[fixupptr]].a:= codeindex;
                        DEC(fixupptr);
                      }
    | if_prefix statement _ELSE
                      {
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                        Generate(jpu,0,0);
                        code[fixup[fixupptr-1]].a:= codeindex;
                      }
      statement
                      {
                        code[fixup[fixupptr]].a:= codeindex;
                        DEC(fixupptr, 2);
                      }
    ;

if_prefix
    : _IF condition _THEN
                      {
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                        Generate(jpc,0,0);
                      }
    ;

for_statement
    : _FOR for_init for_rest
    ;

for_init
    : IDENTIFIER
                      {
                        id:= $1;
                        i:= Position(id,tx[txptr]);
                        IF i=0 THEN
                          Error(11)
                        ELSE IF symtable[i].kind <> varobj THEN
                          Error(12)
                        ELSE
                          WITH symtable[i] DO
                          BEGIN
                            Generate(lit,0,addr);
                            Generate(lda,lev-level,0)
                          END;
                      }
      ASSIGNMENT expression
    | IDENTIFIER LBRACKET
                      {
                        id:= $1;
                        i:= Position(id,tx[txptr]);
                        IF i=0 THEN
                          Error(11)
                        ELSE IF symtable[i].kind <> arrobj THEN
                          Error(49);
                        INC(arrptr);
                        IF arrptr>maxstack THEN Error(32);
                        arr[arrptr]:= i;
                        generate(lit,0,symtable[i].addr);
                      }
      expression RBRACKET
                      {
                        i:= arr[arrptr];
                        generate(add,0,0);
                        generate(lda,lev-symtable[i].level, 0);
                        DEC(arrptr);
                      }
      ASSIGNMENT expression
    ;

for_rest
    : _TO expression _DO
                      {
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                        Generate(feu,0,0);
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                      }
      statement
                      {
                        Generate(fru,0,fixup[fixupptr]);
                        DEC(fixupptr);
                        code[fixup[fixupptr]].a:= codeindex;
                        DEC(fixupptr);
                      }
    | _DOWNTO expression _DO
                      {
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                        Generate(fed,0,0);
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                      }
      statement
                      {
                        Generate(frd,0,fixup[fixupptr]);
                        DEC(fixupptr);
                        code[fixup[fixupptr]].a:= codeindex;
                        DEC(fixupptr);
                      }
    ;

procedure_statement
    : CALL IDENTIFIER
                      {
                        id:= $2;
                        i:= Position(id,tx[txptr]);
                        IF i=0 THEN
                          Error(11)
                        ELSE
                          WITH symtable[i] DO
                          IF kind=procobj THEN
                            Generate(cal,lev-level,addr)
                          ELSE
                            Error(15);
                      }
    ;

read_statement
    : _READ LPAREN IDENTIFIER RPAREN
                      {
                        id:= $3;
                        i:= Position(id,tx[txptr]);
                        IF i=0 THEN
                          Error(11)
                        ELSE IF symtable[i].kind<>varobj THEN
                          Error(12);
                        Generate(inp,0,0);
                        WITH symtable[i] DO
                          Generate(sto,lev-level,addr);
                      }
    ;

repeat_statement
    : _REPEAT
                      {
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                      }
      statement_sequence
      _UNTIL condition
                      {
                        Generate(jpc,0,fixup[fixupptr]);
                        DEC(fixupptr);
                      }
    ;

while_statement
    : _WHILE
                      {
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                      }
      condition
                      {
                        INC(fixupptr);
                        IF fixupptr>maxstack THEN Error(32);
                        fixup[fixupptr]:= codeindex;
                        Generate(jpc,0,0);
                      }
      _DO statement
                      {
                        Generate(jpu,0,fixup[fixupptr-1]);
                        code[fixup[fixupptr]].a:= codeindex;
                        DEC(fixupptr, 2);
                      }
    ;

write_statement
    : _WRITE LPAREN writeparameter_sequence RPAREN
    ;

writeparameter_sequence
    : writeparameter_sequence COMMA writeparameter
    | writeparameter
    ;

writeparameter
    : expression
                      { Generate(prn,0,0); }
    | _STRING
                      {
                        str:= $1;
                        n:= length(str);
                        Generate(prs,0,n);
                        for i:= 1 to n do
                           Generate(asc,0,ord(str[i]));
                      }
    ;

writeln_statement
    : _WRITELN
                      { Generate(prl,0,0); }
    ;

condition
    : ODD expression
                      { Generate(od ,0,0); }
    | expression EQUAL expression
                      { Generate(eq ,0,0); }
    | expression NOTEQUAL expression
                      { Generate(ne ,0,0); }
    | expression LESS expression
                      { Generate(ls ,0,0); }
    | expression GREATER expression
                      { Generate(gt ,0,0); }
    | expression LESSE expression
                      { Generate(le ,0,0); }
    | expression GREATERE expression
                      { Generate(ge ,0,0); }
    ;

expression
    : factor
    | MINUS factor %prec UMINUS
                      { Generate(neg,0,0); }
    | expression PLUS expression
                      { Generate(add,0,0); }
    | expression MINUS expression
                      { Generate(sub,0,0); }
    | expression STAR expression
                      { Generate(mpy,0,0); }
    | expression SLASH expression
                      { Generate(dvd,0,0); }
    ;

factor
    : IDENTIFIER
                      {
                        id:= $1;
                        i:= Position(id,tx[txptr]);
                        IF i=0 THEN
                          Error(11)
                        ELSE
                          WITH symtable[i] DO
                          CASE kind OF
                            constobj: Generate(lit,0,val);
                            varobj  : Generate(lod,lev-level,addr);
                            arrobj  : Error(50);
                            procobj : Error(21);
                          END;
                      }
    | IDENTIFIER LBRACKET
                      {
                        id:= $1;
                        i:= Position(id,tx[txptr]);
                        INC(arrptr);
                        IF arrptr>maxstack THEN Error(32);
                        arr[arrptr]:= i;
                        IF i=0 THEN
                          Error(11)
                        ELSE IF symtable[i].kind <> arrobj THEN
                          Error(49);
                      }
      expression RBRACKET
                      {
                        i:= arr[arrptr];
                        DEC(arrptr);
                        generate(lit,0,symtable[i].addr);
                        generate(add,0,0);
                        generate(drf,lev-symtable[i].level,0);
                      }
    | NUMBER
                      {
                        num:= $1;
                        Generate(lit,0,num);
                      }
    | LPAREN expression RPAREN
    ;

%%

begin
   fixupptr := 0;
   pposptr  := 0;
   dindexptr:= 0;
   caseptr  := 0;
   txptr    := 1;
   arrptr   := 0;
   tx[1]    := 0;
   lev      := 0;

   if (yyparse=0) and (errorsw=no) then
      Writeln('Compilation successfull.')
   else
   begin
      errorsw:= yes;
      WriteLn('Compilation not successfull.');
   end;
   stop;
end.
