{********************************************************************}
{   Program:  LEX/YACC-Compiler for PL/0                             }
{   Unit   :  PL0CODE - Code generation utilities                    }
{   File   :  PL0CODE.PAS                                            }
{   Author :  H.Weber                                                }
{********************************************************************}

UNIT pl0code;

INTERFACE
USES pl0err;

CONST
  codemax = 1000;                 {maximum length of intermediate code          }

TYPE
  opcode = (add, sub, mpy, dvd, neg, od, eq, ne, ls, gt, le, ge,
            lit, ret, cal, lod, sto, Int, jpu, jpc,
            inp, prn, asc, prs, prl, { Erweiterung 18.8.1989 }
            jpx, ldl,             { Erweiterung 21.8.1989 }
            lda, feu, fed,        { Erweiterung  2.1.1990 }
            fru, frd,             { Erweiterung  2.1.1990 }
            sta, drf);            { Erweiterung  8.1.1992 }

  instr = PACKED RECORD
                   o : opcode;    {operation code mnemonic           }
                   l : Integer;   {declaration level                 }
                   a : Integer;   {relative address within DSA       }
                 END;

VAR
  code : ARRAY[0..codemax] OF instr; {intermediate code array }
  codeindex : Integer;            {intermediate code index }
  mnemo : ARRAY[opcode] OF String[5]; {intermed. code mnemonics}
  rcode : Text;                   {output file for intermediate code in    }
  {readable form                           }
  icode : FILE OF instr;
  {output file for intermediate code in    }
  {internal form                           }


PROCEDURE generate(opc : opcode; lv : Integer; Addr : Integer);


IMPLEMENTATION

  PROCEDURE generate(opc : opcode; lv : Integer; Addr : Integer);
    {--------------------------------------------------------------------}
    {   The procedure Generate generates intermediate code statements and}
    {   and writes them out to the codefile.                             }
    {--------------------------------------------------------------------}

  BEGIN                           {Generate}
    IF errorsw = no THEN
    BEGIN
      IF codeindex > codemax THEN BEGIN
        error(31);
        codeindex := codeindex-1;
      END;
      WITH code[codeindex] DO BEGIN
        o := opc;
        l := lv;
        a := Addr
      END;
      codeindex := codeindex+1
    END
  END;                            {Generate}

BEGIN
  codeindex := 0;
END.
