{********************************************************************}
{   Program:  LEX/YACC-Compiler for PL/0                             }
{   Module :  PL0STOP - Stop Routine                                 }
{   File   :  PL0STOP.PAS                                            }
{   Author :  H. Weber / H. Kopp                                     }
{********************************************************************}

UNIT pl0stop;

INTERFACE

USES pl0init, pl0code, pl0err;

PROCEDURE stop;

IMPLEMENTATION

  PROCEDURE stop;
    {--------------------------------------------------------------------}
    {   The procedure Stop completes the listing, writes out intermediate}
    {   code, if existing, closes all files and stops execution.         }
    {--------------------------------------------------------------------}

  VAR errorcount : Integer;
    index : Integer;
    icodename : String;
    rcodename : String;

  BEGIN                           {Stop}

    {print out intermediate code in readable and internal form        }
    {---------------------------------------------------------        }

    icodename := Copy(sourcefile, 1, Pos('.', sourcefile)-1);
    rcodename := icodename+'.cdr';
    icodename := icodename+'.cod';

    Assign(icode, icodename);
    (*$I-*) Rewrite(icode); (*$I+*)
    IF IoResult <> 0 THEN
    BEGIN
      WriteLn('Error opening code file ', icodename);
      Halt(1);
    END;

    Assign(rcode, rcodename);
    (*$I-*) Rewrite(rcode); (*$I+*)
    IF IoResult <> 0 THEN
    BEGIN
      WriteLn('Error opening code file ', rcodename);
      Halt(1);
    END;

    {output in readable form}
    FOR index := 0 TO codeindex-1 DO
      WITH code[index] DO
      BEGIN
        Write(rcode, 'L', index, ':');
        IF index < 10 THEN
          Write(rcode, '     ')
        ELSE IF index < 100 THEN
          Write(rcode, ' ':4)
        ELSE Write(rcode, ' ':3);
        Write(rcode, mnemo[code[index].o]:5);
        IF o IN [inp, prn, asc, prs, prl, feu, fed, fru, frd] THEN
          Write(rcode, a);        { Erw. 2.1.1990 }
        IF o IN [lod, sto, cal] THEN
          Write(rcode, l, ',');
        IF o IN [jpu, cal, jpc] THEN
          Write(rcode, 'L');
        IF o IN [lod, sto, cal, jpu, jpc, lit, int]
        THEN Write(rcode, a);     { Erw. 2.1.1990 }
        IF o IN [jpx, ldl]
        THEN Write(rcode, a);     { Erw. 21.8.1989 }
        IF o IN [lda, sta, drf]
        THEN Write(rcode, l);
        WriteLn(rcode);
      END;
    Close(rcode);
    IF errorsw = yes THEN Erase(rcode);

    {output in internal form}
    FOR index := 0 TO codeindex-1 DO
      Write(icode, code[index]);
    Close(icode);
    IF errorsw = yes THEN Erase(icode);

    IF errorsw = no THEN
      Halt(0)
    ELSE
      Halt(1);
  END {Stop} ;

BEGIN
END.
