{********************************************************************}
{   Program:  LEX/YACC-Compiler for PL/0                             }
{   Unit   :  PL0SYMT - Symbol table manipulation                    }
{   File   :  PL0SYMT.PAS                                            }
{   Author :  H.Weber, H.Kopp                                        }
{********************************************************************}

UNIT pl0symt;

INTERFACE

USES pl0init, pl0err;

CONST
  txmax = 100;                    {maximum value for symbol table index         }

TYPE
  objekt = (constobj, varobj, procobj, arrobj);

VAR
  symtable : ARRAY[0..txmax] OF
    RECORD                        {symbol table}
      name : alfa;
      CASE kind : objekt OF
        constobj : (Val : Integer);
        varobj,
        arrobj,
        procobj : (level, addr, size : Integer)
    END {CASE} ;
  lev, txfirst : Integer;

FUNCTION position(id : alfa; VAR tx : Integer) : Integer;
PROCEDURE enter(k : objekt; VAR dindex, tx : Integer);

IMPLEMENTATION

  FUNCTION position(id : alfa; VAR tx : Integer) : Integer;
    {------------------------------------------------------------------}
    {   The function  Position  searches the position of a given       }
    {   identifier in the symbol table. The returned function value is:}
    {   - the index of the identifier in the symbol table if found or  }
    {   - 0 if the identifier is not found                             }
    {------------------------------------------------------------------}

  VAR i : Integer;

  BEGIN                           {Position}
    symtable[0].name := id;
    i := tx;
    WHILE symtable[i].name <> id DO i := i-1;
    position := i
  END {Position} ;


  PROCEDURE enter(k : objekt; VAR dindex, tx : Integer);
    {------------------------------------------------------------------}
    {   The procedure  Enter  enters an object into the symbol table   }
    {------------------------------------------------------------------}

  VAR i : Integer;                {----- Erweiterung 10.12.1991 -----------}

  BEGIN                           {Enter}
    {----- Beginn Erweiterung 10.12.1991 ----}
    i := position(id, tx);
    IF i >= txfirst THEN error(45);
    {----- Ende Erweiterung 10.12.1991 ------}
    tx := tx+1;
    WITH symtable[tx] DO
    BEGIN
      name := id;
      kind := k;
      CASE k OF constobj : Val := num;
        varobj : BEGIN
                   level := lev;
                   Addr := dindex;
                   dindex := dindex+1
                 END;
        arrobj : BEGIN
                   level := lev;
                   Addr := dindex;
                   dindex := dindex+arrsize;
                   size := arrsize
                 END;
        procobj : level := lev
      END                         {CASE}
    END
  END {Enter} ;

BEGIN
END.

